from github import Github
from github.GithubException import BadCredentialsException as GithubBadCredentialsException
from github.PaginatedList import PaginatedList
from github.AuthenticatedUser import AuthenticatedUser
from github.Organization import Organization
from .feedstockrot import FeedstockRot
import argparse
import os
import logging
from typing import List
from .package import Package


def main() -> int:

    parser = argparse.ArgumentParser(
        description='Check for outdated conda-forge packages'
    )
    parser.add_argument('--debug', action='store_true', help='Enable debug output')

    parser.add_argument(
        '--github',
        action='store_true',
        help='Use your Github repositories as a package list. Requires FEEDSTOCKROT_GITHUB_TOKEN env var'
    )
    # TODO: consider something like this:
    # parser.add_argument(
    #     '--all',
    #     action='store_true',
    #     help='Check all conda-forge packages'
    # )

    parser.add_argument(
        'packages', nargs='*', help='Additional packages to check'
    )

    args = parser.parse_args()

    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    if not args.github and len(args.packages) < 1:
        # well, nothing to do here.
        return 0

    rot = FeedstockRot()

    rot.add(args.packages)
    logging.debug('Added packages from CLI')

    if args.github:
        token = os.getenv('FEEDSTOCKROT_GITHUB_TOKEN', None)
        if not token:
            logging.error('No Github token found')
            return 1

        gh = Github(token)

        try:
            gh_user = gh.get_user()  # type: AuthenticatedUser
            logging.debug('Authenticated to GitHub as: {}'.format(gh_user.name))
        except GithubBadCredentialsException:
            logging.error('Authentication to GitHub failed')
            return 1
        print("Authenticated to GitHub")

        gh_repos = gh_user.get_repos()  # type: PaginatedList
        logging.debug('Got GitHub repositories reference')

        gh_repos = filter(lambda repo: repo.permissions.push, gh_repos)
        logging.debug('Filter repos by push access')

        rot.add_repositories(gh_repos)
        logging.debug('Added repos as packages')

    up_to_date = []  # type: List[Package]
    unknown = []  # type: List[Package]
    upgradeable = []  # type: List[Package]
    not_found = []  # type: List[Package]

    if len(rot.packages) < 1:
        print("No packages")
        return 0

    for pkg in rot.packages:
        if not pkg.latest_feedstock_version:
            not_found.append(pkg)
        elif pkg.latest_external_upgradeable_version:
            upgradeable.append(pkg)
        elif not pkg.latest_external_version:
            unknown.append(pkg)
        else:
            up_to_date.append(pkg)

    if len(up_to_date):
        print("Up-to-date:")
        for pkg in up_to_date:
            print("- {}".format(pkg.name))
    if len(unknown):
        print("Unknown (check these manually):")
        for pkg in unknown:
            print("- {}: {}".format(pkg.name, pkg.latest_feedstock_version))
    if len(upgradeable):
        print("Upgradeable:")
        for pkg in upgradeable:
            # TODO: print out source name if different
            print("- {}: {} -> {}".format(
                pkg.name, pkg.latest_feedstock_version, pkg.latest_external_upgradeable_version))
    if len(not_found):
        print("Not found (no feedstock found, check for typos):")
        for pkg in not_found:
            print("- {}".format(pkg.name))

    return 0


def main_run():
    import sys
    sys.exit(main())


if __name__ == '__main__':
    main_run()



from github.Repository import Repository
from typing import Set, Iterable
from .package import Package


class FeedstockRot:
    def __init__(self, packages: Iterable[str]=None):
        self.packages = set()  # type: Set[Package]

        if packages is not None:
            self.add(packages)

    def add(self, packages: Iterable[str]):
        self.packages |= set(map(lambda v: Package(v), packages))

    def _add_repositories(self, repositories: Iterable[Repository]):
        self.add(map(lambda repo: repo.name, repositories))

    def add_repositories(self, repositories: Iterable[Repository]):
        # TODO: maybe factor out these literals:
        repositories = filter(
            lambda repo:
                repo.owner.login == 'conda-forge' and
                repo.name.endswith('-feedstock'),
            repositories
        )
        self._add_repositories(repositories)



from .source import Source, PackageInfo
from typing import Set, List, Dict
import requests
import yaml
import jinja2


# From https://github.com/conda-forge/conda-smithy/blob/master/conda_smithy/lint_recipe.py#L19
# Allows recipes to be parsed even with undefined jinja2 variables
class Jinja2NullUndefined(jinja2.Undefined):
    def __unicode__(self):
        return str(self._undefined_name)

    def __getattr__(self, name):
        return str('{}.{}'.format(self, name))

    def __getitem__(self, name):
        return '{}["{}"]'.format(self, name)


class Condaforge(Source):

    _DEFAULT_OWNER = 'conda-forge'
    _DEFAULT_PLATFORMS = ['noarch', 'linux-64', 'osx-64', 'win-64']
    _DEFAULT_REPODATA_URL = 'https://conda.anaconda.org/{}/{}/repodata.json'
    _DEFAULT_RECIPE_URL = 'https://raw.githubusercontent.com/conda-forge/{}-feedstock/master/recipe/meta.yaml'

    _repodata = {}

    @classmethod
    def _possible_names(cls, package: PackageInfo):
        name = package.get_name()
        names = list(super()._possible_names(package))
        if name.endswith('-feedstock'):
            names.append(name[:-len('-feedstock')])
        return names

    @classmethod
    def _get_repodata(cls, platform) -> Dict:
        """
        {
            "packages": {
                "$package_name": {
                    "name": "",
                    "version": "",
                    ...
                }
            }
        }
        """
        if platform not in cls._repodata:
            url = cls._DEFAULT_REPODATA_URL.format(cls._DEFAULT_OWNER, platform)
            response = requests.get(url)
            if response.status_code != 200:
                return None

            cls._repodata[platform] = response.json()
        return cls._repodata[platform]

    @classmethod
    def _get_repodata_packages_aggregate(cls) -> List[Dict]:
        packages = []
        for platform in cls._DEFAULT_PLATFORMS:
            repodata = cls._get_repodata(platform)
            if repodata is None:
                continue

            platform_packages = repodata['packages'].values()
            packages += platform_packages
        return packages

    @classmethod
    def _fetch_versions(cls, name: str) -> Set[str]:
        versions = set()
        for package in cls._get_repodata_packages_aggregate():
            if package['name'] == name:
                versions.add(package['version'])
        return versions if len(versions) > 0 else None

    def _get_recipe(self) -> Dict:
        resp = requests.get(self._DEFAULT_RECIPE_URL.format(self.name))
        if resp.status_code != 200:
            return None

        # conda-forge recipes commonly use jinja2 for variables
        try:
            parsed = jinja2.Template(resp.text, undefined=Jinja2NullUndefined)
            rendered = parsed.render()
            return yaml.load(rendered)
        except (jinja2.TemplateError, yaml.YAMLError):
            return None

    def get_recipe_urls(self) -> List[str]:
        """
        Get URLs for a feedstock that may be useful for discerning project info/versions from other sources
        """
        recipe = self._get_recipe()
        urls = []

        if recipe is None:
            return urls

        if 'about' in recipe and 'home' in recipe['about']:
            urls.append(recipe['about']['home'])
        if 'source' in recipe:
            if 'url' in recipe['source']:
                urls.append(recipe['source']['url'])
            if 'git_url' in recipe['source']:
                urls.append(recipe['source']['git_url'])

        return urls



from responses import RequestsMock
from feedstockrot.package_sources.pypi import Pypi
import re
from .mock import Mock
from .mock import mock_package_list
from ..helpers import escape_regex_format_str


class PypiMock(Mock):

    def setup(self, rsps: RequestsMock):
        for package_name in self.package_names:
            response = {"releases": {
                package['version']: {}
                for package in mock_package_list
                if package['name'] == package_name
            }}
            rsps.add(rsps.GET, Pypi.DEFAULT_PACKAGE_URL.format(package_name), json=response)

        re_notfound = re.compile(escape_regex_format_str(Pypi.DEFAULT_PACKAGE_URL).format("([a-zA-Z-_]+)"))
        for _ in range(self._expected_missing):
            rsps.add(rsps.GET, re_notfound, status=404)



from responses import RequestsMock
from typing import Set, List
import abc

mock_package_list = [
    {
        "name": "package_a",
        "version": "1.0"
    },
    {
        "name": "package_a",
        "version": "1.2"
    },
    {
        "name": "package_a",
        "version": "2.0"
    },
    {
        "name": "package_b",
        "version": "1.0",
        "home_url": "https://example.com/package_b/"
    },
    {
        "name": "package_g",
        "version": "1.0",
        "source_url": "https://github.com/package_g/package_g"
    },
    {
        "name": "package_g",
        "version": "1.5",
        "source_url": "https://github.com/package_g/package_g"
    },
    {
        "name": "package_z",
        "version": "1.invalid"
    },
    {
        "name": "package_z",
        "version": "0.1"
    }
]


class Mock(metaclass=abc.ABCMeta):

    def __init__(self, *package_names: List[str]):
        self.package_names = set(package_names)
        self._expected_missing = 0

    def add(self, package_name):
        self.package_names.add(package_name)
        return self

    def expected_missing(self, expected_missing: int):
        self._expected_missing = expected_missing
        return self

    def __enter__(self):
        pass

    def __exit__(self, *exc):
        pass

    @abc.abstractmethod
    def setup(self, rsps: RequestsMock):
        pass


class Mocker(object):

    def __init__(self, *mockers: List[Mock]):
        self.rsps = RequestsMock()
        self.mockers = mockers

    def add(self, *package_names: str):
        for package_name in package_names:
            for mock in self.mockers:
                mock.add(package_name)

    def __enter__(self):
        self.rsps.reset()
        for mocker in self.mockers:
            mocker.setup(self.rsps)

        for mocker in self.mockers:
            mocker.__enter__()

        self.rsps.__enter__()

    def __exit__(self, *exc):
        self.rsps.__exit__(*exc)

        for mocker in self.mockers:
            mocker.__exit__(*exc)



from unittest import TestCase
from feedstockrot.package_sources.pypi import Pypi
from feedstockrot.package import Package
from packaging.version import Version
from ..helpers.mock.mock import Mocker
from ..helpers.mock.pypi import PypiMock
from ..helpers.packageinfo import PackageInfoFake


class TestPypi(TestCase):

    def test_possible_names(self):
        name = 'testing'
        package_names = {
            '{}-python'.format(name), '{}-py'.format(name),
            'python-{}'.format(name), 'py-{}'.format(name)
        }

        for package_name in package_names:
            possible = Pypi._possible_names(PackageInfoFake(package_name))
            self.assertListEqual(
                [package_name, name],
                possible
            )

    def test_fetch_versions(self):
        pkg_name = 'package_a'

        with Mocker(PypiMock(pkg_name)):
            result = Pypi._fetch_versions(pkg_name)

        self.assertIsNotNone(result)

    def test_versions(self):
        pkg = Package('package_a')
        with Mocker(PypiMock(pkg.name)):
            src = Pypi(pkg)
            result = src.versions

        self.assertSetEqual({Version('1.0'), Version('1.2'), Version('2.0')}, result)

        pkg = Package('package_z')
        with Mocker(PypiMock(pkg.name)):
            src = Pypi(pkg)
            result = src.versions

        self.assertSetEqual({Version('0.1')}, result)



from .source import Source, PackageInfo
import requests
from typing import Dict, Set, List
import logging
from urllib.parse import urlparse
from pathlib import PurePosixPath
from os.path import splitext

class Github(Source):

    # For a Github object the `name` is owner/repo as a single string
    # so the URLs below are formatted with that assumption

    _INFO_URL = 'https://api.github.com/repos/{}'
    _RELEASES_URL = 'https://api.github.com/repos/{}/releases'
    _REFS_URL = 'https://api.github.com/repos/{}/git/refs'
    _TAGS_URL = 'https://api.github.com/repos/{}/git/refs/tags'
    _TAG_PREFIX = 'refs/tags/'

    @classmethod
    def _possible_names(cls, package: PackageInfo) -> List[str]:
        names = list()
        for url_str in package.get_urls():
            url = urlparse(url_str)
            host = url.netloc
            if not host.endswith('github.com'):
                continue

            path = PurePosixPath(url.path)
            parts = path.parts
            if path.is_absolute():
                parts = parts[1:]

            if len(parts) >= 2:
                # Get the first 2 path components without extensions
                # this should handle:
                # - owner/project
                # - owner/project.git
                # - owner/project/releases

                name = PurePosixPath(parts[0])
                # strip off .git if the project name contains it
                # don't just strip off any ext because "." is valid
                name_project = parts[1]
                if name_project.endswith('.git'):
                    name_project = name_project[:-len('.git')]
                name = name.joinpath(name_project)

                names.append(str(name))

        return names

    @classmethod
    def _fetch_tags(cls, name: str) -> Dict:
        response = requests.get(cls._REFS_URL.format(name))
        if response.status_code != 200:
            return None
        tags_data = response.json()

        versions = set()
        for tag_data in tags_data:
            ref = tag_data['ref']
            if ref.startswith(cls._TAG_PREFIX):
                tag = ref[len(cls._TAG_PREFIX):]
                versions.add(tag)
        return versions

    @classmethod
    def _fetch_versions(cls, name: str) -> Set[str]:
        return cls._fetch_tags(name)



from unittest import TestCase
from feedstockrot.package_sources.github import Github
from feedstockrot.package import Package
from packaging.version import Version
from ..helpers.mock.mock import Mocker
from ..helpers.mock.github import GithubMock
from ..helpers.packageinfo import PackageInfoFake


class TestGithub(TestCase):

    def test_possible_names(self):
        name = 'testing'
        urls = [
            'https://github.com/testowner/testproj',
            'https://github.com/testowner/testproj.git',
            'https://github.com/testowner/testproj/releases',
        ]
        possible_names = ['testowner/testproj']

        for url in urls:
            possible = Github._possible_names(PackageInfoFake(name, urls=[url]))
            self.assertListEqual(
                possible_names,
                possible
            )

    def test_fetch_versions(self):
        with Mocker(GithubMock('package_a')):
            result = Github._fetch_versions('package_a/package_a')

        self.assertIsNotNone(result)

    def test_versions(self):
        pkg = PackageInfoFake('package_g', ['https://github.com/package_g/package_g'])
        with Mocker(GithubMock(pkg.name)):
            src = Github(pkg)
            result = src.versions

        self.assertSetEqual({Version('1.0'), Version('1.5')}, result)

        pkg = PackageInfoFake('package_z', ['https://github.com/package_z/package_z'])
        with Mocker(GithubMock(pkg.name)):
            src = Github(pkg)
            result = src.versions

        self.assertSetEqual({Version('0.1')}, result)



from .mock import Mock
from .mock import mock_package_list
from responses import RequestsMock
from feedstockrot.package_sources.condaforge import Condaforge
import yaml


class CondaforgeRepoMock(Mock):

    def __enter__(self):
        # Reset the repodata after each test so there is no dependency on it being cached
        self._old_repodata = Condaforge._repodata
        Condaforge._repodata = {}

        # Make things simpler by only dealing with 1 platform for now:
        self._old_platforms = Condaforge._DEFAULT_PLATFORMS
        Condaforge._DEFAULT_PLATFORMS = [Condaforge._DEFAULT_PLATFORMS[0]]

    def __exit__(self, *exc):
        Condaforge._repodata = self._old_repodata
        Condaforge._DEFAULT_PLATFORMS = self._old_platforms

    def setup(self, rsps: RequestsMock):
        response = {"packages": {
            "{}-{}".format(package['name'], package['version']): package
            for package in mock_package_list
            if package['name'] in self.package_names
        }}
        rsps.add(
            rsps.GET,
            Condaforge._DEFAULT_REPODATA_URL.format(Condaforge._DEFAULT_OWNER, Condaforge._DEFAULT_PLATFORMS[0]),
            json=response
        )


class CondaforgeRecipeMock(Mock):

    def setup(self, rsps: RequestsMock):
        for package_name in self.package_names:
            package = [package for package in mock_package_list if package['name'] == package_name][-1]
            recipe = {"package": {"name": package_name, "version": package['version']}}
            if "source_url" in package:
                recipe["source"] = {"url": package['source_url']}
            if "home_url" in package:
                recipe['about'] = {"home": package['home_url']}
            response = yaml.dump(recipe)
            rsps.add(rsps.GET, Condaforge._DEFAULT_RECIPE_URL.format(package_name), body=response)



from unittest import TestCase
from feedstockrot.feedstockrot import FeedstockRot
from feedstockrot.package import Package
from github.Repository import Repository
from .helpers.mock.mock import Mocker
from .helpers.mock.condaforge import CondaforgeRepoMock
from .helpers.mock.pypi import PypiMock
from packaging.version import Version
from feedstockrot.package_sources.pypi import Pypi


class TestFeedstockrot(TestCase):

    def setUp(self):
        # use a single source here just to ensure functionality
        self._old_sources = Package._SOURCE_CLASSES
        Package._SOURCE_CLASSES = [Pypi]

        self.rot = FeedstockRot()

    def tearDown(self):
        Package._SOURCE_CLASSES = self._old_sources

    def test_add(self):
        packages = {'package_a', 'package_b'}
        with Mocker(CondaforgeRepoMock(*packages), PypiMock(*packages)):
            self.rot.add(packages)

        self.assertEqual(len(packages), len(self.rot.packages))
        rot_packages = self.rot.packages.copy()

        for package in packages:
            rot_package = next(pkg for pkg in rot_packages if pkg.name == package)
            self.assertIsNotNone(rot_package)
            self.assertIsInstance(rot_package, Package)

    def test_add_repositories(self):
        repositories_good = [
            Repository(None, None, {"name": "package_a-feedstock", "owner": {"login": "conda-forge"}}, True),
            Repository(None, None, {"name": "package_b-feedstock", "owner": {"login": "conda-forge"}}, True),
        ]
        repositories_bad_owner = [
            Repository(None, None, {"name": "package_a-feedstock", "owner": {"login": "johnsmith"}}, True),
            Repository(None, None, {"name": "package_c-feedstock", "owner": {"login": "joesmith"}}, True),
        ]
        repositories_bad_name = [
            Repository(None, None, {"name": "package_d", "owner": {"login": "conda-forge"}}, True),
        ]
        repositories_bad_name_owner = [
            Repository(None, None, {"name": "package_e", "owner": {"login": "conda-forge"}}, True),
        ]

        repositories = repositories_good + repositories_bad_owner + repositories_bad_name + repositories_bad_name_owner
        package_names = {'package_a', 'package_b', 'package_c', 'package_d', 'package_e'}
        with Mocker(CondaforgeRepoMock(*package_names), PypiMock().expected_missing(2)):
            self.rot.add_repositories(repositories)
        rot_packages = self.rot.packages.copy()

        self.assertEqual(len(repositories_good), len(rot_packages))

        for pkg in rot_packages:  # type: Package
            if pkg.get_name() == 'package_a':
                self.assertEqual(Version('2.0'), pkg.latest_feedstock_version)
                self.assertIsNone(pkg.latest_external_version) # we're responding with a 404 to pypi
                self.assertIsNone(pkg.latest_external_upgradeable_version)
            elif pkg.get_name() == 'package_b':
                self.assertEqual(Version('1.0'), pkg.latest_feedstock_version)
                self.assertIsNone(pkg.latest_external_version) # we're responding with a 404 to pypi
                self.assertIsNone(pkg.latest_external_upgradeable_version)
            else:
                self.assertIsNone(pkg.get_name())



from responses import RequestsMock
from feedstockrot.package_sources.github import Github
import re
from .mock import Mock
from .mock import mock_package_list
from ..helpers import escape_regex_format_str


class GithubMock(Mock):

    def setup(self, rsps: RequestsMock):
        for package_name in self.package_names:
            response = [
                {"ref": "refs/tags/{}".format(package['version'])}
                for package in mock_package_list
                if package['name'] == package_name
            ]

            github_name = package_name
            if '/' not in github_name:
                github_name = package_name + "/" + package_name

            rsps.add(rsps.GET, Github._REFS_URL.format(github_name), json=response)

        re_notfound = re.compile(escape_regex_format_str(Github._REFS_URL).format("([a-zA-Z-_.]+)"))
        for _ in range(self._expected_missing):
            rsps.add(rsps.GET, re_notfound, status=404)



from .package_sources.condaforge import Condaforge
from .package_sources.pypi import Pypi
from .package_sources.github import Github
from .package_sources.source import PackageInfo
import itertools


def value_or_empty_set(value):
    if value is None:
        return set()
    return value


class Package(PackageInfo):

    # not including condaforge:
    _SOURCE_CLASSES = [Pypi, Github]

    def __init__(self, name):
        self.name = name

        self._source_condaforge = Condaforge(self)
        self._sources_external = {}
        for source_cls in self._SOURCE_CLASSES:
            self._sources_external[source_cls] = source_cls(self)

    @property
    def versions_condaforge(self):
        return value_or_empty_set(self._source_condaforge.versions)

    @property
    def _external_versions(self):
        return itertools.chain.from_iterable(
            value_or_empty_set(source.versions)
            for source in self._sources_external.values()
        )

    @property
    def _external_upgradeable_versions(self):
        versions = self._external_versions

        # Only newer versions:
        versions = filter(lambda v: v > self.latest_feedstock_version, versions)

        # Only early versions if we're already on one:
        if not self.latest_feedstock_version or not self.latest_feedstock_version.is_prerelease:
            versions = filter(lambda v: not v.is_prerelease, versions)

        return versions

    @property
    def latest_feedstock_version(self):
        return max(self.versions_condaforge, default=None)

    @property
    def latest_external_version(self):
        return max(self._external_versions, default=None)

    @property
    def latest_external_upgradeable_version(self):
        return max(self._external_upgradeable_versions, default=None)

    def __str__(self):
        return "<{} condaforge={} external={} upgradeable={}>".format(
            self.name, self.latest_feedstock_version,
            self.latest_external_version, self.latest_external_upgradeable_version
        )

    def get_name(self):
        # check required for when we actually call Condaforge(self) before it is set
        if hasattr(self, '_source_condaforge') and self._source_condaforge.name is not None:
            return self._source_condaforge.name
        return self.name

    def get_urls(self):
        return self._source_condaforge.get_recipe_urls()



import abc
from typing import Iterable, Set
from packaging.version import Version, InvalidVersion
import logging


class PackageInfo(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get_name(self) -> str:
        pass

    @abc.abstractmethod
    def get_urls(self) -> Iterable[str]:
        pass


class Source(metaclass=abc.ABCMeta):

    def __init__(self, package: PackageInfo):
        self.package = package

        self.name = None
        self._data_versions = None
        self._versions = None

        # Find source-specific name:
        for name in self._possible_names(self.package):
            data = self._fetch_versions(name)
            if data is not None:
                self._data_versions = data
                self.name = name
                break

    @classmethod
    @abc.abstractmethod
    def _fetch_versions(cls, name: str) -> Set[str]:
        pass

    @classmethod
    def _possible_names(cls, package: PackageInfo) -> Iterable[str]:
        """
        Return some iterable (list is preferred) of possible package names
        """
        return {package.get_name()}

    @property
    def versions(self) -> Set[Version]:
        if self._versions is not None:
            return self._versions
        if self._data_versions is None:
            return None

        versions = set()
        for version_str in self._data_versions:
            try:
                version = Version(version_str)
            except InvalidVersion:
                logging.info("Got invalid version for {}: {}".format(self.package.get_name(), version_str))
                continue
            versions.add(version)

        self._versions = versions
        return self._versions



from feedstockrot.package_sources.source import PackageInfo


class PackageInfoFake(PackageInfo):
    def __init__(self, name, urls=None):
        self.name = name
        self.urls = urls if urls else []

    def get_name(self):
        return self.name

    def get_urls(self):
        return self.urls



from unittest import TestCase
from feedstockrot.package_sources.condaforge import Condaforge
from feedstockrot.package import Package
from packaging.version import Version
from ..helpers.mock.mock import Mocker
from ..helpers.mock.condaforge import CondaforgeRepoMock
from ..helpers.mock.condaforge import CondaforgeRecipeMock
from ..helpers.packageinfo import PackageInfoFake


class TestCondaforge(TestCase):

    def setUp(self):
        self._old_sources = Package._SOURCE_CLASSES
        Package._SOURCE_CLASSES = []

    def tearDown(self):
        Package._SOURCE_CLASSES = self._old_sources

    def test_possible_names(self):
        self.assertListEqual(
            ['testing-feedstock', 'testing'],
            list(Condaforge._possible_names(PackageInfoFake('testing-feedstock')))
        )

    # TODO: consider testing multi-platform capability
    def test_get_repodata(self):
        with Mocker(CondaforgeRepoMock('package_a')):
            result = Condaforge._get_repodata(Condaforge._DEFAULT_PLATFORMS[0])

        self.assertIsNotNone(result)
        self.assertIsInstance(result, dict)
        self.assertIn('packages', result)
        self.assertIn('package_a-1.0', result['packages'])

    def test_fetch_versions(self):
        with Mocker(CondaforgeRepoMock('package_a')):
            result_a = Condaforge._fetch_versions("package_a")
        expected_a = {'1.0', '1.2', '2.0'}
        self.assertSetEqual(result_a, expected_a)

        with Mocker(CondaforgeRepoMock('package_b')):
            result_b = Condaforge._fetch_versions("package_b")
        self.assertSetEqual(result_b, {'1.0'})

    def test_versions(self):
        with Mocker(CondaforgeRepoMock('package_a')):
            src_a = Condaforge(Package('package_a'))

        self.assertSetEqual(
            {Version('1.0'), Version('1.2'), Version('2.0')},
            src_a.versions
        )

        with Mocker(CondaforgeRepoMock('package_b')):
            src_b = Condaforge(Package('package_b'))

        self.assertSetEqual(
            {Version('1.0')},
            src_b.versions
        )

        with Mocker(CondaforgeRepoMock('package_z')):
            src_z = Condaforge(Package('package_z'))

        self.assertSetEqual(
            {Version('0.1')},
            src_z.versions
        )

    def test__get_recipe_a(self):
        with Mocker(CondaforgeRepoMock('package_a')):
            src_a = Condaforge(Package('package_a'))
        with Mocker(CondaforgeRecipeMock('package_a')):
            result = src_a._get_recipe()
        self.assertIsInstance(result, dict)
        self.assertIn('package', result)

    def test__get_recipe_b(self):
        with Mocker(CondaforgeRepoMock('package_b')):
            src_a = Condaforge(Package('package_b'))
        with Mocker(CondaforgeRecipeMock('package_b')):
            result = src_a._get_recipe()
        self.assertIsInstance(result, dict)
        self.assertIn('package', result)
        self.assertIn('about', result)
        self.assertIn('home', result['about'])

    def test_get_recipe_urls_a(self):
        with Mocker(CondaforgeRepoMock('package_a')):
            src_a = Condaforge(Package('package_a'))
        with Mocker(CondaforgeRecipeMock('package_a')):
            result = src_a.get_recipe_urls()
        self.assertEqual(0, len(result))

    def test_get_recipe_urls_b(self):
        with Mocker(CondaforgeRepoMock('package_b')):
            src_a = Condaforge(Package('package_b'))
        with Mocker(CondaforgeRecipeMock('package_b')):
            result = src_a.get_recipe_urls()
        self.assertEqual(1, len(result))



from .source import Source, PackageInfo
import requests
from typing import Dict, Set, List
import logging


class Pypi(Source):

    DEFAULT_PACKAGE_URL = "https://pypi.python.org/pypi/{}/json"

    @classmethod
    def _possible_names(cls, package: PackageInfo) -> List[str]:
        names = list(super()._possible_names(package))
        name = package.get_name()

        if name.startswith('python-'):
            names.append(name[len('python-'):])
        elif name.startswith('py-'):
            names.append(name[len('py-'):])

        if name.endswith('-python'):
            names.append(name[:-len('-python')])
        elif name.endswith('-py'):
            names.append(name[:-len('-py')])

        if len(names) > 1:
            logging.debug('Possible names for {}: {}'.format(name, names))

        return names

    @classmethod
    def _fetch(cls, name) -> Dict:
        resp = requests.get(cls.DEFAULT_PACKAGE_URL.format(name))
        if resp.status_code != 200:
            return None
        return resp.json()

    @classmethod
    def _fetch_versions(cls, name: str) -> Set[str]:
        resp = cls._fetch(name)
        if resp is None:
            return None
        return resp['releases'].keys()



from unittest import TestCase
from feedstockrot.package import Package, value_or_empty_set
from packaging.version import Version
from .helpers.mock.mock import Mocker
from .helpers.mock.pypi import PypiMock
from .helpers.mock.condaforge import CondaforgeRepoMock
from feedstockrot.package_sources.pypi import Pypi


class TestPackage(TestCase):

    def setUp(self):
        # use a single source here, just to ensure the functionality works
        self._old_sources = Package._SOURCE_CLASSES
        Package._SOURCE_CLASSES = [Pypi]

        with Mocker(PypiMock('package_a'), CondaforgeRepoMock('package_a')):
            self.pkg_a = Package('package_a')
        with Mocker(PypiMock().expected_missing(1), CondaforgeRepoMock()):
            self.pkg_bad = Package('not-a-real-package')

    def tearDown(self):
        Package._SOURCE_CLASSES = self._old_sources

    def test_value_or_empty_set(self):
        self.assertSetEqual(set(), value_or_empty_set(None))
        self.assertSetEqual(set(), value_or_empty_set(set()))
        self.assertSetEqual({'test'}, value_or_empty_set({'test'}))
        self.assertEqual(['test'], value_or_empty_set(['test']))

    def test_versions_condaforge(self):
        self.assertSetEqual({Version('1.0'), Version('1.2'), Version('2.0')}, self.pkg_a.versions_condaforge)
        self.assertSetEqual(set(), self.pkg_bad.versions_condaforge)

    def test__external_versions(self):
        self.assertSetEqual({Version('1.0'), Version('1.2'), Version('2.0')}, set(self.pkg_a._external_versions))
        self.assertSetEqual(set(), set(self.pkg_bad._external_versions))

    def test__external_upgradeable_versions(self):
        class Patch(Package):
            versions_condaforge = {Version('1.0'), Version('1.2')}  # missing 2.0
        self.pkg_a.__class__ = Patch

        self.assertSetEqual({Version('2.0')}, set(self.pkg_a._external_upgradeable_versions))

        # TODO: test other cases, including dev versions

    def test_latest_feedstock_version(self):
        self.assertEqual(Version('2.0'), self.pkg_a.latest_feedstock_version)
        self.assertIsNone(self.pkg_bad.latest_feedstock_version)

    def test_latest_external_version(self):
        self.assertEqual(Version('2.0'), self.pkg_a.latest_external_version)
        self.assertIsNone(self.pkg_bad.latest_external_version)

    def test_latest_external_upgradeable_version(self):
        class Patch(Package):
            versions_condaforge = {Version('1.0'), Version('1.2')}  # missing 2.0
        self.pkg_a.__class__ = Patch

        self.assertEqual(Version('2.0'), self.pkg_a.latest_external_upgradeable_version)

        self.assertIsNone(self.pkg_bad.latest_external_upgradeable_version)

    def test_get_name(self):
        self.assertEqual(self.pkg_a.name, self.pkg_a.get_name())
        self.assertEqual(self.pkg_bad.name, self.pkg_bad.get_name())

    # duplicate of condaforge test
    # def test_get_urls(self):
